;; Generated package description from aggressive-indent.el  -*- no-byte-compile: t -*-
(define-package "aggressive-indent" "1.10.0.0.20210701.192409" "Minor mode to aggressively keep your code always indented" '((emacs "24.3")) :authors '(("Artur Malabarba" . "emacs@endlessparentheses.com")) :maintainer '("Artur Malabarba" . "emacs@endlessparentheses.com") :keywords '("indent" "lisp" "maint" "tools") :url "https://github.com/Malabarba/aggressive-indent-mode")
